 /*
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM.  Reproduction, modification and redistribution are prohibited.
 * PageModel.js
 * (c) Copyright IBM Corporation 2011, 2014.  
 * U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
 
dojo.provide("com.ibm.rdm.web.grid.PageModel");
 
dojo.require("com.ibm.rdm.web.common.MimeTypes");
dojo.require("com.ibm.rdm.web.notifications.ChangeNotifier");
 // to get default page size.
dojo.require("com.ibm.rdm.web.artifact.Constants");

(function () { 


	// Summary:
	//   Divides the module structure into pages.
	//   Each page has a list of the URIs that should be on it.
	

	var Constants = com.ibm.rdm.web.artifact.Constants,
		MimeTypeRegistry = com.ibm.rdm.web.common.MimeTypeRegistry,
		ChangeNotifier = com.ibm.rdm.web.notifications.ChangeNotifier,
		ROW_HEIGHT_ESTIMATE = 50;
	
	var CACHE = {};

	dojo.declare("com.ibm.rdm.web.grid.PageModel", null, {
	
		constructor: function (args) {
			dojo.mixin(this, args);
			this._heights = {};
			this._imgHeights = {};
			this._images = {};
			this._structureLoadedConnection = dojo.connect(this.structure, "onLoad", this, function (args) { 
				this.onChange(args);
			});
			this._structureChangedConnection = dojo.connect(this.structure, "onChange", this, function (args) { 
				this.onChange(args);
			});
			
			if (this.isStructureLoaded()) {
				this.buildPages();
			}
			
			
			this._saveListener = ChangeNotifier.addListener({
				eventTopics: [
					com.ibm.rdm.web.notifications.EventTopics.ARTIFACT_SAVED
					],
				callbackFunction: dojo.hitch(this, this._updateOnSave),
				callbackScope: this,
				projectURI: this.structure.projectURI
			});
			
			
			this._samplePageIndex = 0;
			CACHE[this.structure.moduleURI] = this;
			
		},
		
		isStructureLoaded: function () {
			return this.structure.isLoaded;
		},
		
		cleanup: function () {
			dojo.disconnect(this._structureLoadedConnection);
			dojo.disconnect(this._structureChangedConnection);
			ChangeNotifier.removeListener(this._saveListener);
			delete CACHE[this.structure.moduleURI];
		},
		
		setDisplayImages: function (display) {
			this._displayImages = display;
		},
		
		getDisplayImages: function () {
			return this._displayImages;
		},
		
		addImage: function (/*String*/uri, /*Element*/imgNode) {
			// Summary:
			//  Adds an image (lazily loaded probably) into the
			//  model for the given URI.  This is currently used
			//  for RRCx artifacts, where we do not want to preload
			//  convertor images so instead we wait until the user
			//  has loaded the page they are on, then call this
			//  method to update the model.
			//  This allows the fisheye to display a scaled version
			//  of the image, once we have loaded it.
			
			if (!imgNode) {
			    return;
			}
			
			var clone = imgNode.cloneNode(true);
			this._images[uri] = clone;
			
			if (imgNode.complete) {
				this._imgHeights[uri] = imgNode.offsetHeight;
				if (!dojo.isFF) {
					clone.width = imgNode.offsetWidth;
				}
			} else {
				imgNode.onload = dojo.hitch(this, function() {
					this._imgHeights[uri] = imgNode.offsetHeight;
					delete this._heights[uri];
					if (!dojo.isFF) {
						clone.width = imgNode.offsetWidth;
					}
				});
			}
			
		},
		
		getImage: function (uri) {
			return this._images[uri];
		},
		
		_findImages: function () {
			var images = dojo.filter(this.allArtifacts, function (entry) {
				return entry.wrappedResourceMimeType && MimeTypeRegistry.isImage(entry.wrappedResourceMimeType);
			});
			
			var imageReady = dojo.hitch(this, function (img, entry) {
				this._imgHeights[entry.uri] = img.height;
				this._images[entry.uri] = img;
			});
			
			dojo.forEach(images, dojo.hitch(this, function (entry) {
				var img = new Image();
				
				img.onload = function () {
					imageReady(img,entry);
				};
				
				img.src = entry.wrappedResourceURI+"?accept=none";
				if (img.complete) {
					imageReady(img,entry);
				}
				
			}));
		},
		
		onChange: function (args) {
			this.buildPages();
			this._findImages();
			this.calculateHeightAndPositions(true);
		},
		
		setHeight: function (uri, height) {
			this._heights[uri] = height;
		},
		
		getHeight: function (uri) {
			return (this._displayImages && this._imgHeights[uri]) || this._heights[uri] || ROW_HEIGHT_ESTIMATE;
		},
		
		getImageWidth: function (uri) {
			var img = this._images[uri];
			// if the image hasn't loaded, then just return the estimate of the row height,
			// we have no idea what the image dimensions are anyway.
			return ((img && (img.width || img.offsetWidth)) || ROW_HEIGHT_ESTIMATE);
		},
		
		clearHeights: function () {
			this._heights = {};
		},

		buildPages: function () {
		
			// TODO: should we put a layer of abstraction between the structure and the page model ?
			var artifacts = this.structure.getArtifacts(),
				length = artifacts.length,
				numRowsPerPage = Constants.DEFAULT_PAGE_SIZE,
				numPages = Math.ceil(length /numRowsPerPage),
				pageY = 0,
				startRow,
				endRow,
				page;
        
        	this.pages = [];
        
        	for (var i = 0; i < numPages; i++) {
				startRow = i * numRowsPerPage;
                endRow = Math.min((i * numRowsPerPage) + (numRowsPerPage), length);
            	page = {
					pageNum: (i+1),
                	y: pageY,
                	startRow: startRow,
                	endRow: endRow,
                	heightEstimated: true,
					artifacts: artifacts.slice(startRow, endRow),
					uris: []
            	};
				
				var idx, len = page.artifacts.length;
					
				for (idx=0; idx<len; idx++) {
					page.uris.push(page.artifacts[idx].uri);
				}

            	pageY += page.height;
				
				this.pages.push(page);
			}

			if (this.pages.length > 0) {
				this.pages[numPages - 1].isLast = true;
			}	
			
			// cache this - since calling getArtifacts will filter based on
			// expanded or collapsed ancestory 
			this.allArtifacts = artifacts;
		},
		
		getPreviousURI: function (uri) {
			// Summary:
			//  given the supplied URI, what entry comes
			//  immediately before it in the list ?
			var entry = this.structure.getArtifactByURI(uri),
				idx,
				artifacts = this.allArtifacts;
			
			if (entry) {
				idx = dojo.indexOf(artifacts, entry);
				if (idx > 0 && idx < artifacts.length) {
					return artifacts[idx-1].uri;
				}
			}
		},
		
		getNextURI: function (uri) {
			// Summary:
			//  given the supplied URI, what entry comes
			//  immediately after it in the list ?
			var entry = this.structure.getArtifactByURI(uri),
				idx, 
				artifacts = this.allArtifacts;
			
			if (entry) {
				idx = dojo.indexOf(artifacts, entry);
				if (idx < artifacts.length-2) {
					return artifacts[idx+1].uri;
				}
			}
		},
		
		getPageForScrollPosition: function (scrollTop) {
			var page, 			// the page from this.pages[] that matches the scroll pos
			topOfPage = 0;  // counter for the top of the current page in loop to calc. correct page.
			
			for (var i = 0; i < this.pages.length && !page; i++) {
            	var p = this.pages[i];
            	// does the scroll index fall within the
            	// bounds of this page ?
            	if (topOfPage + p.height >= scrollTop) {
                	page = p;
            	}
			
            	if (!page) {
                	topOfPage += p.height;
            	}
        	}

			return page || this.pages[this.pages.length-1];
		},
		
		getPage: function (idx) {
			return this.pages[idx];
		},
		
		getPages: function () {
			return this.pages;
		},
		
		getPageForURI: function (uri) {
			for (var i=0; i<this.pages.length; i++) {
				var p = this.pages[i];
				if (dojo.indexOf(p.uris, uri) >= 0) {
					return p;
				}
			}	
		},
		
		getEarlierRowAtOffset: function(rowKey, offset) {
			// summary:
			//		Find the row that is the given offset above the given row.
			// rowKey:
			//		key (URI) identifying the row to work from.
			// offset:
			//		distance in pixels of our desired target position.
			// returns:
			//		The key (URI) of a row, or undefined if there is no
			//		available row above the given row.
			
			// Find the page and row position of the reference row.
			var pageIndex,
				rowIndex = -1,
				page,
				pageCount = this.pages.length;
			for (pageIndex = 0; pageIndex < pageCount ; pageIndex++) {
				page = this.pages[pageIndex];
				if ((rowIndex = dojo.indexOf(page.uris, rowKey)) >= 0) {
					break;
				}
			}
			if (rowIndex < 0) {
				// no pages - empty document, or for some other reason we didn't find
				// what we were looking for.
				return;
			}
			
			// Walk up through the rows to find the one we want, based on height.
			// We iterate this in such a way that we will always move at least one
			// row upwards.
			while (true) {
				if (--rowIndex < 0) {
					// change page and set rowIndex to last row on that page
					if (pageIndex <= 0) {
						// no more pages
						break;
					}
					page = this.pages[--pageIndex];
					rowIndex = page.rowHeights.length - 1;
				}
				offset -= page.rowHeights[rowIndex];
				if (offset <= 0) {
					// we are at (or, more likely, slightly above) our target postion
					return page.uris[rowIndex];
				}
			}

			// got to the end without getting to our offset, so return the
			// first row, unless our starting point was already the first row.
			var firstKey = this.pages[0].uris[0];
			if (rowKey !== firstKey) {
				return firstKey;
			}
			// implicitly return undefined - we are already at the first row. 
		},
		
		getLaterRowAtOffset: function(rowKey, offset) {
			// summary:
			//		Find the row that is the given offset below the given row.
			// rowKey:
			//		key (URI) identifying the row to work from.
			// offset:
			//		distance in pixels of our desired target position.
			// returns:
			//		The key (URI) of a row, or undefined if there is no
			//		available row below the given row.
			
			// Find the page and row position of the reference row.
			var pageIndex,
				rowIndex = -1,
				page,
				pageCount = this.pages.length;
			for (pageIndex = 0; pageIndex < pageCount ; pageIndex++) {
				page = this.pages[pageIndex];
				if ((rowIndex = dojo.indexOf(page.uris, rowKey)) >= 0) {
					break;
				}
			}
			if (rowIndex < 0) {
				// no pages - empty document, or for some other reason we didn't find
				// what we were looking for.
				return;
			}
			
			// Walk down through the rows to find the one we want, based on height.
			// We iterate this in such a way that we will always move at least one
			// row downwards.  Take into account the height of the current row, but
			// *not* the target row, so giving the same measures as getEarlierRowAtOffset().
			// This should make page up/page down operations move corresponding
			// amounts.
			var targetIsNextRow;
			while (true) {
				offset -= page.rowHeights[rowIndex];
				if (offset <= 0) {
					// we are at (or, more likely, slightly below) our target postion
					targetIsNextRow = true;
				}
				if (++rowIndex >= page.rowHeights.length) {
					// change page and set rowIndex to first row on that page
					if (++pageIndex >= pageCount) {
						// no more pages
						break;
					}
					page = this.pages[pageIndex];
					rowIndex = 0;
				}
				if (targetIsNextRow) {
					return page.uris[rowIndex];
				}
			}

			// got to the end without getting to our offset, so return the
			// last row, unless our starting point was already the last row.
			var uris = this.pages[pageCount-1].uris;
			var lastKey = uris[uris.length-1];
			if (rowKey !== lastKey) {
				return lastKey;
			}
			// implicitly return undefined - we are already at the last row. 
		},
		
		calculateHeightAndPositions: function(estimated) {
		//var start = new Date().getTime();			
			// Summary:
			//   assigns a new height to the DOM node containing the virtual
			//   grid, thus adjusting the size of the scroll track.  Simply
			//   runs over each page and totals the height.  if we have pages
			//   that have an estimated height, then their height is based on
			//   the height of an average row on the page index indicated as 
			//   _samplePageIndex.
	        var totalHeight = 0, 
				page, 
				pageY = 0,
				oldHeight;
	        
	        // Fire event to remember the current top visible item and scroll it back to the view later if necessary
	        ChangeNotifier.notifyListeners({
				eventTopic: com.ibm.rdm.web.notifications.EventTopics.GRID_PRE_PAGE_HEIGHT_RECALCULATE
		    });
	        
	        for (var i = 0; i < this.pages.length; i++) {
	            page = this.pages[i];
	            if (!page.heightEstimated) {
	            	oldHeight = page.height;
	            }
				page.height = 0;
				page.rowHeights = [];
				var len = page.uris.length;
				for (var idx = 0; idx < len; idx++) {
					var uri = page.uris[idx];
					var height = this.getHeight(uri);
					page.height += height;
					page.rowHeights.push(height);
				}

	            totalHeight += page.height;
	            if (!page.heightEstimated) {
		            if (oldHeight !== page.height) {
		            	ChangeNotifier.notifyListeners({
							eventTopic: com.ibm.rdm.web.notifications.EventTopics.GRID_PAGE_HEIGHT_CHANGE
						});
		            }
	            }
	            
	            page.y = pageY;
	            pageY += page.height;
	        }

	        this.totalHeight = totalHeight;
			//console.debug("calcHeightAndPositions took ", new Date().getTime() - start);				
    	},
	
		getTotalHeight: function () {
			return this.totalHeight;
		},
		
		getArtifactsAsList: function () {
			return this.allArtifacts;
		},
		
		indexOf: function (/*String*/uri) {
			// Summary:
			//   return the index of the given URI in the model.
			//	 If the given artifact is collapsed/not showing, index will not be returned.
			var entry = this.structure.getArtifactByURI(uri),
				idx, 
				artifacts = this.allArtifacts;
			
			if (entry) {
				idx = artifacts.indexOf(entry);
			}
			return idx;
		},
		
		_updateStubRowOnSave: function(/*String*/preSaveURI, /*AbstractResource*/resource) {
			// Summary:
			//
			//   Called when an object is saved so that if the page model contained
			//   a "stub" entry pre save, then that entry is replaced by the proper
			//   URI.  
			//
			
			var i,
				pageCount,
			 	idx,
			 	page;
			
			if (preSaveURI && preSaveURI.indexOf("STUB") === 0) {
				for (i=0, pageCount = this.pages.length; i < pageCount; i++) {
					page = this.pages[i];
					idx = page.uris.indexOf(preSaveURI);
					if (idx !== -1) {
						page.uris[idx] = resource.resourceURI;
					}
				}
			}
		},
		
		_updateOnSave: function(args) {
			if (args && args.notifierParams && args.notifierParams.resourceURI) {
				//Find and swap new properties from newly created/saved resource
				this._updateStubRowOnSave(args.notifierParams.resourceURI, args.notifierParams.resources[0]);
			} else if (args && args.notifierParams && args.notifierParams.resources) {
				dojo.forEach(args.notifierParams.resources, function(resource) {
					this._updateStubRowOnSave(resource.resourceURI, resource);
				});
			}
		}
	
	});
	
	com.ibm.rdm.web.grid.PageModel.getCachedModel = function (moduleURI) {
		return CACHE[moduleURI];
	};

} ());
